@@ -44,6 +44,7 @@ import (
 	"github.com/containerd/containerd/runtime/v2/runc/options"
 	"github.com/containerd/containerd/runtime/v2/shim"
 	taskAPI "github.com/containerd/containerd/runtime/v2/task"
+	"github.com/containerd/containerd/sys/reaper"
 	runcC "github.com/containerd/go-runc"
 	"github.com/containerd/typeurl"
 	ptypes "github.com/gogo/protobuf/types"
@@ -78,12 +79,12 @@ func New(ctx context.Context, id string, publisher events.Publisher) (shim.Shim,
 		context:   ctx,
 		processes: make(map[string]rproc.Process),
 		events:    make(chan interface{}, 128),
-		ec:        shim.Default.Subscribe(),
+		ec:        reaper.Default.Subscribe(),
 		ep:        ep,
 		cancel:    cancel,
 	}
 	go s.processExits()
-	runcC.Monitor = shim.Default
+	runcC.Monitor = reaper.Default
 	if err := s.initPlatform(); err != nil {
 		cancel()
 		return nil, errors.Wrap(err, "failed to initialized platform behavior")
